@@ -1,6 +1,6 @@
 # sdtfile.py
 
-# Copyright (c) 2007-2022, Christoph Gohlke
+# Copyright (c) 2007-2024, Christoph Gohlke
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -41,87 +41,64 @@ equipment for photon counting.
 
 :Author: `Christoph Gohlke <https://www.cgohlke.com>`_
 :License: BSD 3-Clause
-:Version: 2022.9.28
+:Version: 2024.11.24
+:DOI: `10.5281/zenodo.10125608 <https://doi.org/10.5281/zenodo.10125608>`_
+
+Quickstart
+----------
+
+Install the sdtfile package and all dependencies from the
+`Python Package Index <https://pypi.org/project/sdtfile/>`_::
+
+    python -m pip install -U sdtfile
+
+See `Examples`_ for using the programming interface.
+
+Source code and support are available on
+`GitHub <https://github.com/cgohlke/sdtfile>`_.
 
 Requirements
 ------------
 
-This release has been tested with the following requirements and dependencies
+This revision was tested with the following requirements and dependencies
 (other versions may work):
 
-- `CPython 3.8.10, 3.9.13, 3.10.7, 3.11.0rc2 <https://www.python.org>`_
-- `Numpy 1.22.4 <https://pypi.org/project/numpy/>`_
+- `CPython <https://www.python.org>`_ 3.10.11, 3.11.9, 3.12.7, 3.13.0
+- `Numpy <https://pypi.org/project/numpy/>`_ 2.1.3
 
 Revisions
 ---------
 
-2022.9.28
+2024.11.24
 
-- Convert docstrings to Google style with Sphinx directives.
+- Update MEASURE_INFO_EXT struct to SPCM v.9.9 (breaking).
 
-2022.2.2
+2024.5.24
 
-- Add type hints.
-- Drop support for Python 3.7 and numpy < 1.19 (NEP29).
+- Fix docstring examples not correctly rendered on GitHub.
 
-2021.11.18
+2024.4.24
 
-- Fix reading FLIM files created by Prairie View software (#5).
+- Support NumPy 2.
 
-2021.3.21
+2023.9.28
 
-- Add sdt2dat script.
+- Update structs to SPCM v.9.66 (breaking).
+- Shorten MEASURE_INFO struct to meas_desc_block_length.
 
-2020.12.10
+2023.8.30
 
-- Fix shape of non-square frames.
+- …
 
-2020.8.3
-
-- Fix integer overflow (#3).
-- Support os.PathLike file names.
-
-2020.1.1
-
-- Fix reading MCS_BLOCK data.
-- Remove support for Python 2.7 and 3.5.
-- Update copyright.
-
-2019.7.28
-
-- Fix reading compressed, multi-channel data.
-
-2018.9.22
-
-- Use str, not bytes for ASCII data.
-
-2018.8.29
-
-- Move module into sdtfile package.
-
-2018.2.7
-
-- Bug fixes.
-
-2016.3.30
-
-- Support revision 15 files and compression.
-
-2015.1.29
-
-- Read SPC DLL data files.
-
-2014.9.5
-
-- Fix reading multiple MEASURE_INFO records.
+Refer to the CHANGES file for older revisions.
 
 References
 ----------
 
-1. W Becker. The bh TCSPC Handbook. Third Edition. Becker & Hickl GmbH 2008.
-   pp 401.
+1. W Becker. The bh TCSPC Handbook. 9th Edition. Becker & Hickl GmbH 2021.
+   pp 879.
 2. SPC_data_file_structure.h header file. Part of the Becker & Hickl
-   SPCM software.
+   SPCM software installation.
 
 Examples
 --------
@@ -129,11 +106,11 @@ Examples
 Read image and metadata from a "SPC Setup & Data File":
 
 >>> sdt = SdtFile('image.sdt')
->>> sdt.header.revision
+>>> int(sdt.header.revision)
 588
 >>> sdt.info.id[1:-1]
 'SPC Setup & Data File'
->>> int(sdt.measure_info[0].scan_x)
+>>> int(sdt.measure_info[0].scan_x[0])
 128
 >>> len(sdt.data)
 1
@@ -142,7 +119,7 @@ Read image and metadata from a "SPC Setup & Data File":
 >>> sdt.times[0].shape
 (256,)
 
-Read data and metadata from a "SPC Setup & Data File" with mutliple data sets:
+Read data and metadata from a "SPC Setup & Data File" with multiple data sets:
 
 >>> sdt = SdtFile('fluorescein.sdt')
 >>> len(sdt.data)
@@ -168,9 +145,10 @@ Read image data from a "SPC FCS Data File" as numpy array:
 
 from __future__ import annotations
 
-__version__ = '2022.9.28'
+__version__ = '2024.11.24'
 
 __all__ = [
+    '__version__',
     'SdtFile',
     'FileInfo',
     'SetupBlock',
@@ -179,13 +157,18 @@ __all__ = [
     'FileRevision',
 ]
 
-import os
 import io
+import os
 import zipfile
-from typing import BinaryIO
+from typing import TYPE_CHECKING
 
 import numpy
 
+if TYPE_CHECKING:
+    from typing import Any, BinaryIO
+
+    from numpy.typing import NDArray
+
 
 class SdtFile:
     """Becker & Hickl SDT file.
@@ -197,22 +180,29 @@ class SdtFile:
 
     filename: str
     """Name of file."""
-    header: numpy.recarray
+
+    header: numpy.recarray[Any, Any]
     """File header of type FILE_HEADER."""
+
     info: FileInfo
     """File info string and attributes."""
+
     setup: SetupBlock | None
     """Setup block ascii and binary data."""
-    measure_info: list[numpy.recarray]
+
+    measure_info: list[numpy.recarray[Any, Any]]
     """Measurement description blocks of type MEASURE_INFO."""
-    block_headers: list[numpy.recarray]
+
+    block_headers: list[numpy.recarray[Any, Any]]
     """Data block headers of type BLOCK_HEADER."""
-    data: list[numpy.ndarray]
+
+    data: list[NDArray[Any]]
     """Photon counts at each curve point."""
-    times: list[numpy.ndarray]
+
+    times: list[NDArray[Any]]
     """Time axes for each data set."""
 
-    def __init__(self, arg: str | os.PathLike | BinaryIO, /) -> None:
+    def __init__(self, arg: str | os.PathLike[Any] | BinaryIO, /) -> None:
         if isinstance(arg, (str, os.PathLike)):
             self.filename = os.fspath(arg)
             with open(arg, 'rb') as fh:
@@ -225,10 +215,14 @@ class SdtFile:
     def _fromfile(self, fh: BinaryIO, /) -> None:
         """Initialize instance from open file."""
         # read file header
-        self.header = numpy.rec.fromfile(  # type: ignore
-            fh, dtype=FILE_HEADER, shape=1, byteorder='<'
+        self.header = numpy.rec.fromfile(
+            # type: ignore[call-overload]
+            fh,
+            dtype=FILE_HEADER,
+            shape=1,
+            byteorder='<',
         )[0]
-        if self.header.header_valid != 0x5555:
+        if self.header.chksum != 0x55AA and self.header.header_valid != 0x5555:
             raise ValueError('not a SDT file')
         if self.header.no_of_data_blocks == 0x7FFF:
             self.header.no_of_data_blocks = self.header.reserved1
@@ -236,7 +230,7 @@ class SdtFile:
             raise ValueError('')
 
         # read file info
-        fh.seek(self.header.info_offset)
+        fh.seek(self.header.info_offs)
         info = fh.read(self.header.info_length).decode('windows-1250')
         info = info.replace('\r\n', '\n')
         self.info = FileInfo(info)
@@ -261,32 +255,42 @@ class SdtFile:
 
         # read measurement description blocks
         self.measure_info = []
-        dtype = numpy.dtype(MEASURE_INFO)
-        if dtype.itemsize > self.header.meas_desc_block_length:
-            # TODO: shorten MEASURE_INFO to meas_desc_block_length
-            pass
-        fh.seek(self.header.meas_desc_block_offset)
+        dtype = struct_dtype(
+            MEASURE_INFO, int(self.header.meas_desc_block_length)
+        )
+        fh.seek(self.header.meas_desc_block_offs)
         for _ in range(self.header.no_of_meas_desc_blocks):
             self.measure_info.append(
-                numpy.rec.fromfile(  # type: ignore
-                    fh, dtype=dtype, shape=1, byteorder='<'
+                numpy.rec.fromfile(
+                    # type: ignore[call-overload]
+                    fh,
+                    dtype=dtype,
+                    shape=1,
+                    byteorder='<',
                 )
             )
             fh.seek(self.header.meas_desc_block_length - dtype.itemsize, 1)
 
         rev = FileRevision(self.header.revision)
-        block_header_t = BLOCK_HEADER if rev.revision < 15 else BLOCK_HEADER_15
+        if rev.revision >= 15:
+            block_header_t = BLOCK_HEADER
+        else:
+            block_header_t = BLOCK_HEADER_OLD
 
         self.times = []
         self.data = []
         self.block_headers = []
 
-        offset = self.header.data_block_offset
+        offset = self.header.data_block_offs
         for _ in range(self.header.no_of_data_blocks):
             # read data block header
             fh.seek(offset)
-            bh = numpy.rec.fromfile(  # type: ignore
-                fh, dtype=block_header_t, shape=1, byteorder='<'
+            bh = numpy.rec.fromfile(
+                # type: ignore[call-overload]
+                fh,
+                dtype=block_header_t,
+                shape=1,
+                byteorder='<',
             )[0]
             self.block_headers.append(bh)
             # read data block
@@ -305,50 +309,75 @@ class SdtFile:
                 data = numpy.fromfile(fh, dtype=dtype, count=dsize)
 
             # TODO: support more block types
-            # the following works with DECAY_BLOCK, IMG_BLOCK, and MCS_BLOCK
-            adc_re = int(mi.adc_re)
-            scan_x = int(mi.scan_x)
-            scan_y = int(mi.scan_y)
-            image_x = int(mi.image_x)
-            image_y = int(mi.image_y)
+            # the following works with DECAY, IMG, MCS, PAGE
+
+            # assume adc_re is always present
+            adc_re = int(mi.adc_re[0])
+
+            # the following fields may not be present
+            try:
+                scan_x = int(mi.scan_x[0])
+                scan_y = int(mi.scan_y[0])
+            except AttributeError:
+                scan_x = 0
+                scan_y = 0
+            try:
+                image_x = int(mi.image_x[0])
+                image_y = int(mi.image_y[0])
+            except AttributeError:
+                image_x = 0
+                image_y = 0
+            try:
+                mcs_points = mi.MeasHISTInfo.mcs_points[0]
+            except AttributeError:
+                mcs_points = -1
+            try:
+                mcs_time = mi.MeasHISTInfo.mcs_time[0]
+            except AttributeError:
+                mcs_time = 0
+
+            if adc_re == 0:
+                adc_re = 65536
             if dsize == scan_x * scan_y * adc_re:
                 data = data.reshape(scan_y, scan_x, adc_re)
             elif dsize == image_x * image_y * adc_re:
                 data = data.reshape(image_y, image_x, adc_re)
-            elif dsize == mi.MeasHISTInfo.mcs_points[0]:
+            elif dsize == mcs_points:
                 data = data.reshape(-1, dsize)
             else:
                 data = data.reshape(-1, adc_re)
             self.data.append(data)
 
-            if bt.contents == 'MCS_BLOCK':
-                t = numpy.arange(dsize, dtype=numpy.float64)
-                t *= mi.MeasHISTInfo.mcs_time[0]
+            if bt.contents == 'MCS_BLOCK' and mcs_time != 0:
+                time = numpy.arange(dsize, dtype=numpy.float64)
+                time *= mcs_time
             else:
                 # generate time axis
-                t = numpy.arange(adc_re, dtype=numpy.float64)
-                t *= mi.tac_r / (float(mi.tac_g) * adc_re)
-            self.times.append(t)
+                time = numpy.arange(adc_re, dtype=numpy.float64)
+                time *= mi.tac_r / (float(mi.tac_g[0]) * adc_re)
+            self.times.append(time)
             offset = bh.next_block_offs
 
-    def block_measure_info(self, block: int, /) -> numpy.recarray:
+    def block_measure_info(self, block: int, /) -> numpy.recarray[Any, Any]:
         """Return measure_info record for data block.
 
         Parameters:
             block: Block index.
 
         """
-        return self.measure_info[self.block_headers[block].meas_desc_block_no]
+        return self.measure_info[
+            int(self.block_headers[block].meas_desc_block_no)
+        ]
 
     def __enter__(self) -> SdtFile:
         return self
 
-    def __exit__(self, exc_type, exc_value, traceback):
+    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
         pass
 
     def __repr__(self) -> str:
         filename = os.path.split(self.filename)[-1]
-        return f'<{self.__class__.__name__} {filename!r}>'
+        return f'{self.__class__.__name__}({filename!r})'
 
     def __str__(self) -> str:
         return indent(
@@ -371,8 +400,7 @@ class FileInfo(str):
     """File info string and attributes.
 
     Parameters:
-        value:
-            File content from FILE_HEADER info_offset and info_length.
+        value: File content from FILE_HEADER info_offs and info_length.
 
     """
 
@@ -398,8 +426,7 @@ class SetupBlock:
     """Setup block ascii and binary data.
 
     Parameters:
-        value:
-            File content from FILE_HEADER setup_offs and setup_length.
+        value: File content from FILE_HEADER setup_offs and setup_length.
 
     """
 
@@ -407,6 +434,7 @@ class SetupBlock:
 
     ascii: str
     """ASCII data."""
+
     binary: bytes | None
     """Binary data."""
 
@@ -415,7 +443,7 @@ class SetupBlock:
         i = value.find(b'BIN_PARA_BEGIN')
         if i:
             self.ascii = value[:i].decode('windows-1250')
-            self.binary = bytes(value[i + 15 : -10])
+            self.binary = value[i:]  # [i + 15 : -10]
             # TODO: parse binary data here
         else:
             self.ascii = value.decode('windows-1250')
@@ -437,6 +465,7 @@ class BlockNo:
 
     data: int
     """Data."""
+
     module: int
     """Module."""
 
@@ -445,7 +474,7 @@ class BlockNo:
         self.module = value & 0x000000FF
 
     def __repr__(self) -> str:
-        return f'<{self.__class__.__name__} {self.data} {self.module}>'
+        return f'{self.__class__.__name__}({self.data} << 24 & {self.module})'
 
 
 class BlockType:
@@ -460,10 +489,13 @@ class BlockType:
 
     mode: str
     """BLOCK_CREATION."""
+
     contents: str
     """BLOCK_CONTENT."""
-    dtype: numpy.dtype
+
+    dtype: numpy.dtype[Any]
     """BLOCK_DTYPE."""
+
     compress: bool
     """Data is compressed."""
 
@@ -498,6 +530,7 @@ class FileRevision:
 
     revision: int
     """Revision."""
+
     module: str
     """Module."""
 
@@ -515,6 +548,21 @@ class FileRevision:
             0x28: 'SPC-150',
             0x29: 'DPC-230',
             0x2A: 'SPC-130EM',
+            0x2B: 'SPC-160',
+            0x2E: 'SPC-150N',
+            0x80: 'SPC-150NX',
+            0x81: 'SPC-160X',
+            0x82: 'SPC-160PCIE',
+            0x83: 'SPC-130EMN',
+            0x84: 'SPC-180N',
+            0x85: 'SPC-180NX',
+            0x86: 'SPC-180NXX',
+            0x87: 'SPC-180N-USB',
+            0x88: 'SPC-130IN',
+            0x89: 'SPC-130INX',
+            0x8A: 'SPC-130INXX',
+            0x8B: 'SPC-QC-104',
+            0x8C: 'SPC-QC-004',
         }.get((value & 0xFF0) >> 4, 'Unknown')
 
     def __repr__(self) -> str:
@@ -525,14 +573,14 @@ class FileRevision:
 
 FILE_HEADER: list[tuple[str, str]] = [
     ('revision', 'i2'),
-    ('info_offset', 'i4'),
+    ('info_offs', 'i4'),
     ('info_length', 'i2'),
     ('setup_offs', 'i4'),
-    ('setup_length', 'i2'),
-    ('data_block_offset', 'i4'),
+    ('setup_length', 'u2'),
+    ('data_block_offs', 'i4'),
     ('no_of_data_blocks', 'i2'),
-    ('data_block_length', 'i4'),
-    ('meas_desc_block_offset', 'i4'),
+    ('data_block_length', 'u4'),
+    ('meas_desc_block_offs', 'i4'),
     ('no_of_meas_desc_blocks', 'i2'),
     ('meas_desc_block_length', 'i2'),
     ('header_valid', 'u2'),
@@ -593,13 +641,50 @@ HIST_INFO: list[tuple[str, str]] = [
     ('filda_points', 'i4'),
     ('mcs_time', 'f4'),
     ('mcs_points', 'i4'),
+    ('cross_calc_phot', 'u4'),
+    ('mcsta_points', 'u2'),
+    ('mcsta_flags', 'u2'),
+    ('mcsta_tpp', 'u4'),
+    ('calc_markers', 'u4'),
+    ('fcs_calc_phot', 'u4'),
+    ('reserved3', 'u4'),
+]
+
+HIST_INFO_EXT: list[tuple[str, str]] = [
+    ('first_frame_time', 'f4'),
+    ('frame_time', 'f4'),
+    ('line_time', 'f4'),
+    ('pixel_time', 'f4'),
+    ('scan_type', 'i2'),
+    ('skip_2nd_line_clk', 'i2'),
+    ('right_border', 'u4'),
+    ('info', 'S40'),
+]
+
+MEASURE_INFO_EXT: list[tuple[str, str]] = [
+    ('DCU_in_use', 'u4'),
+    ('dcu_ser_no', '4S16'),
+    ('scope_name', 'S32'),
+    ('lens_name', 'S64'),
+    ('SIS_in_use', 'u4'),
+    ('sis_ser_no', '4S16'),
+    ('gvd_ser_no', 'S16'),
+    ('zoom_factor', 'f4'),
+    ('FOV_at_zoom_1', 'f4'),
+    ('scope_connected', 'i2'),
+    ('lens_magnifier', 'f4'),
+    ('image_size', 'f4'),
+    ('tdc_offset', '4f4'),
+    ('tdc_control', 'u4'),
+    ('scale_bar', 'u1'),
+    ('reserve', 'S1249'),
 ]
 
 # Measurement description blocks
 MEASURE_INFO: list[tuple[str, str | list[tuple[str, str]]]] = [
-    ('time', 'a9'),
-    ('date', 'a11'),
-    ('mod_ser_no', 'a16'),
+    ('time', 'S9'),
+    ('date', 'S11'),
+    ('mod_ser_no', 'S16'),
     ('meas_mode', 'i2'),
     ('cfd_ll', 'f4'),
     ('cfd_lh', 'f4'),
@@ -628,7 +713,7 @@ MEASURE_INFO: list[tuple[str, str | list[tuple[str, str]]]] = [
     ('dither', 'i2'),
     ('incr', 'i2'),
     ('mem_bank', 'i2'),
-    ('mod', 'a16'),
+    ('mod', 'S16'),
     ('syn_th', 'f4'),
     ('dead_time_comp', 'i2'),
     ('polarity_l', 'i2'),
@@ -666,9 +751,33 @@ MEASURE_INFO: list[tuple[str, str | list[tuple[str, str]]]] = [
     ('det', 'i2'),
     ('x_axis', 'i2'),
     ('MeasHISTInfo', HIST_INFO),
+    ('HISTInfoExt', HIST_INFO_EXT),
+    ('sync_delay', 'f4'),
+    ('sdel_ser_no', 'u2'),
+    ('sdel_input', 'i1'),
+    ('mosaic_ctrl', 'i1'),
+    ('mosaic_x', 'u1'),
+    ('mosaic_y', 'u1'),
+    ('frames_per_el', 'i2'),
+    ('chan_per_el', 'i2'),
+    ('mosaic_cycles_done', 'i4'),
+    ('mla_ser_no', 'u2'),
+    ('DCC_in_use', 'u1'),
+    ('dcc_ser_no', 'S12'),
+    ('TiSaLas_status', 'u2'),
+    ('TiSaLas_wav', 'u2'),
+    ('AOM_status', 'u1'),
+    ('AOM_power', 'u1'),
+    ('ddg_ser_no', 'S8'),
+    ('prior_ser_no', 'i4'),
+    ('mosaic_x_hi', 'u1'),
+    ('mosaic_y_hi', 'u1'),
+    ('reserve', 'S11'),
+    ('extension_used', 'u1'),
+    ('minfo_ext', MEASURE_INFO_EXT),
 ]
 
-BLOCK_HEADER: list[tuple[str, str]] = [
+BLOCK_HEADER_OLD: list[tuple[str, str]] = [
     ('block_no', 'i2'),
     ('data_offs', 'i4'),
     ('next_block_offs', 'i4'),
@@ -678,7 +787,7 @@ BLOCK_HEADER: list[tuple[str, str]] = [
     ('block_length', 'u4'),
 ]
 
-BLOCK_HEADER_15: list[tuple[str, str]] = [
+BLOCK_HEADER: list[tuple[str, str]] = [
     ('data_offs_ext', 'u1'),
     ('next_block_offs_ext', 'u1'),
     ('data_offs', 'u4'),
@@ -718,7 +827,7 @@ BLOCK_CONTENT: dict[int, str] = {
 }
 
 # Data type
-BLOCK_DTYPE: dict[int, numpy.dtype] = {
+BLOCK_DTYPE: dict[int, numpy.dtype[Any]] = {
     0x000: numpy.dtype('<u2'),
     0x100: numpy.dtype('<u4'),
     0x200: numpy.dtype('<f8'),
@@ -738,7 +847,34 @@ INFO_IDS: dict[str, str] = {
 }
 
 
-def indent(*args) -> str:
+def struct_dtype(
+    struct: list[tuple[str, str | list[tuple[str, str]]]], size: int, /
+) -> numpy.dtype[Any]:
+    """Return numpy dtype for struct not exceeding size bytes."""
+    assert size > 0
+    fields = len(struct)
+    dtype = numpy.dtype(struct)
+    while dtype.itemsize > size and fields > 0:
+        # last_dtype = struct[-1][1]
+        # if (
+        #     isinstance(last_dtype, list)
+        #     and dtype.itemsize - size < numpy.dtype(last_dtype).itemsize
+        # ):
+        #     struct = struct.copy()
+        #     struct[-1] = (
+        #         struct[-1][0],
+        #         struct_dtype(last_dtype, dtype.itemsize - size).descr
+        #     )
+        #     dtype = numpy.dtype(struct)
+        #     break
+        fields -= 1
+        dtype = numpy.dtype(struct[:fields])
+    # if dtype.itemsize != size:
+    #    log_warning(f'struct size {dtype.itemsize} != {size}')
+    return dtype
+
+
+def indent(*args: Any) -> str:
     """Return joined string representations of objects with indented lines."""
     text = '\n'.join(str(arg) for arg in args)
     return '\n'.join(
@@ -746,7 +882,23 @@ def indent(*args) -> str:
     )[2:]
 
 
+def log_warning(msg: object, *args: object, **kwargs: Any) -> None:
+    """Log message with level WARNING."""
+    import logging
+
+    logging.getLogger(__name__).warning(msg, *args, **kwargs)
+
+
 if __name__ == '__main__':
     import doctest
 
     doctest.testmod()
+
+    assert numpy.dtype(FILE_HEADER).itemsize == 42  # BH_HDR_LENGTH
+    assert numpy.dtype(MEASURE_INFO).itemsize == 2048
+    assert numpy.dtype(MEASURE_INFO_EXT).itemsize == 1536
+    assert numpy.dtype(HIST_INFO).itemsize == 48
+    assert numpy.dtype(HIST_INFO_EXT).itemsize == 64
+    assert numpy.dtype(MEASURE_FCS_INFO).itemsize == 38
+    assert numpy.dtype(MEASURE_STOP_INFO).itemsize == 60
+    assert numpy.dtype(SETUP_BIN_HDR).itemsize == 14
